/*
 * (C) Copyright 2007
 * Gerald Van Baren, Custom IDEAS, vanbaren@cideas.com
 * Based on code written by:
 *   Pantelis Antoniou <pantelis.antoniou@gmail.com> and
 *   Matthew McClintock <msm@freescale.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <asm/global_data.h>
#include <fdt.h>
#include <libfdt.h>
#include <fdt_support.h>
#include <ufdt_overlay.h>
#include <asm/arch/system.h>
#include <asm/arch/fw_info.h>

#define MAX_LEVEL	32		/* how deeply nested we will go */
#define SCRATCHPAD	1024		/* bytes of scratchpad memory */

#if defined(CONFIG_DTB_COMPAT_CMDLINE_EXTEND)
#define do_extend_cmdline 1
#else
#define do_extend_cmdline 0
#endif

/*
 * Global data (for the gd->bd)
 */
DECLARE_GLOBAL_DATA_PTR;

static int fdt_valid(void);
static int fdt_parse_prop(char *const*newval, int count, char *data, int *len);
static int fdt_print(const char *pathp, char *prop, int depth);

/*
 * The working_fdt points to our working flattened device tree.
 */
struct fdt_header *working_fdt;

void set_working_fdt_addr(void *addr)
{
	char buf[17];

	working_fdt = addr;

	sprintf(buf, "%lx", (unsigned long)addr);
	setenv("fdtaddr", buf);
}

/*
 * Flattened Device Tree command, see the help for parameter definitions.
 */
int do_fdt (cmd_tbl_t * cmdtp, int flag, int argc, char * const argv[])
{
	if (argc < 2)
		return CMD_RET_USAGE;

	/*
	 * Set the address of the fdt
	 */
	if (argv[1][0] == 'a') {
		unsigned long addr;
		/*
		 * Set the address [and length] of the fdt.
		 */
		if (argc == 2) {
			if (!fdt_valid()) {
				return 1;
			}
			printf("The address of the fdt is %p\n", working_fdt);
			return 0;
		}

		addr = simple_strtoul(argv[2], NULL, 16);
		set_working_fdt_addr((void *)addr);

		if (!fdt_valid()) {
			return 1;
		}

		if (argc >= 4) {
			int  len;
			int  err;
			/*
			 * Optional new length
			 */
			len = simple_strtoul(argv[3], NULL, 16);
			if (len < fdt_totalsize(working_fdt)) {
				printf ("New length %d < existing length %d, "
					"ignoring.\n",
					len, fdt_totalsize(working_fdt));
			} else {
				/*
				 * Open in place with a new length.
				 */
				err = fdt_open_into(working_fdt, working_fdt, len);
				if (err != 0) {
					printf ("libfdt fdt_open_into(): %s\n",
						fdt_strerror(err));
				}
			}
		}

	/*
	 * Move the working_fdt
	 */
	} else if (strncmp(argv[1], "mo", 2) == 0) {
		struct fdt_header *newaddr;
		int  len;
		int  err;

		if (argc < 4)
			return CMD_RET_USAGE;

		/*
		 * Set the address and length of the fdt.
		 */
		working_fdt = (struct fdt_header *)simple_strtoul(argv[2], NULL, 16);
		if (!fdt_valid()) {
			return 1;
		}

		newaddr = (struct fdt_header *)simple_strtoul(argv[3],NULL,16);

		/*
		 * If the user specifies a length, use that.  Otherwise use the
		 * current length.
		 */
		if (argc <= 4) {
			len = fdt_totalsize(working_fdt);
		} else {
			len = simple_strtoul(argv[4], NULL, 16);
			if (len < fdt_totalsize(working_fdt)) {
				printf ("New length 0x%X < existing length "
					"0x%X, aborting.\n",
					len, fdt_totalsize(working_fdt));
				return 1;
			}
		}

		/*
		 * Copy to the new location.
		 */
		err = fdt_open_into(working_fdt, newaddr, len);
		if (err != 0) {
			printf ("libfdt fdt_open_into(): %s\n",
				fdt_strerror(err));
			return 1;
		}
		working_fdt = newaddr;

	/*
	 * Make a new node
	 */
	} else if (strncmp(argv[1], "mk", 2) == 0) {
		char *pathp;		/* path */
		char *nodep;		/* new node to add */
		int  nodeoffset;	/* node offset from libfdt */
		int  err;

		/*
		 * Parameters: Node path, new node to be appended to the path.
		 */
		if (argc < 4)
			return CMD_RET_USAGE;

		pathp = argv[2];
		nodep = argv[3];

		nodeoffset = fdt_path_offset (working_fdt, pathp);
		if (nodeoffset < 0) {
			/*
			 * Not found or something else bad happened.
			 */
			printf ("libfdt fdt_path_offset() returned %s\n",
				fdt_strerror(nodeoffset));
			return 1;
		}
		err = fdt_add_subnode(working_fdt, nodeoffset, nodep);
		if (err < 0) {
			printf ("libfdt fdt_add_subnode(): %s\n",
				fdt_strerror(err));
			return 1;
		}

	/*
	 * Set the value of a property in the working_fdt.
	 */
	} else if (argv[1][0] == 's') {
		char *pathp;		/* path */
		char *prop;		/* property */
		int  nodeoffset;	/* node offset from libfdt */
		static char data[SCRATCHPAD];	/* storage for the property */
		int  len;		/* new length of the property */
		int  ret;		/* return value */

		/*
		 * Parameters: Node path, property, optional value.
		 */
		if (argc < 4)
			return CMD_RET_USAGE;

		pathp  = argv[2];
		prop   = argv[3];
		if (argc == 4) {
			len = 0;
		} else {
			ret = fdt_parse_prop(&argv[4], argc - 4, data, &len);
			if (ret != 0)
				return ret;
		}

		nodeoffset = fdt_path_offset (working_fdt, pathp);
		if (nodeoffset < 0) {
			/*
			 * Not found or something else bad happened.
			 */
			printf ("libfdt fdt_path_offset() returned %s\n",
				fdt_strerror(nodeoffset));
			return 1;
		}

		ret = fdt_setprop(working_fdt, nodeoffset, prop, data, len);
		if (ret < 0) {
			printf ("libfdt fdt_setprop(): %s\n", fdt_strerror(ret));
			return 1;
		}

	/*
	 * Print (recursive) / List (single level)
	 */
	} else if ((argv[1][0] == 'p') || (argv[1][0] == 'l')) {
		int depth = MAX_LEVEL;	/* how deep to print */
		char *pathp;		/* path */
		char *prop;		/* property */
		int  ret;		/* return value */
		static char root[2] = "/";

		/*
		 * list is an alias for print, but limited to 1 level
		 */
		if (argv[1][0] == 'l') {
			depth = 1;
		}

		/*
		 * Get the starting path.  The root node is an oddball,
		 * the offset is zero and has no name.
		 */
		if (argc == 2)
			pathp = root;
		else
			pathp = argv[2];
		if (argc > 3)
			prop = argv[3];
		else
			prop = NULL;

		ret = fdt_print(pathp, prop, depth);
		if (ret != 0)
			return ret;

	/*
	 * Remove a property/node
	 */
	} else if (strncmp(argv[1], "rm", 2) == 0) {
		int  nodeoffset;	/* node offset from libfdt */
		int  err;

		/*
		 * Get the path.  The root node is an oddball, the offset
		 * is zero and has no name.
		 */
		nodeoffset = fdt_path_offset (working_fdt, argv[2]);
		if (nodeoffset < 0) {
			/*
			 * Not found or something else bad happened.
			 */
			printf ("libfdt fdt_path_offset() returned %s\n",
				fdt_strerror(nodeoffset));
			return 1;
		}
		/*
		 * Do the delete.  A fourth parameter means delete a property,
		 * otherwise delete the node.
		 */
		if (argc > 3) {
			err = fdt_delprop(working_fdt, nodeoffset, argv[3]);
			if (err < 0) {
				printf("libfdt fdt_delprop():  %s\n",
					fdt_strerror(err));
				return err;
			}
		} else {
			err = fdt_del_node(working_fdt, nodeoffset);
			if (err < 0) {
				printf("libfdt fdt_del_node():  %s\n",
					fdt_strerror(err));
				return err;
			}
		}

	/*
	 * Display header info
	 */
	} else if (argv[1][0] == 'h') {
		u32 version = fdt_version(working_fdt);
		printf("magic:\t\t\t0x%x\n", fdt_magic(working_fdt));
		printf("totalsize:\t\t0x%x (%d)\n", fdt_totalsize(working_fdt),
		       fdt_totalsize(working_fdt));
		printf("off_dt_struct:\t\t0x%x\n",
		       fdt_off_dt_struct(working_fdt));
		printf("off_dt_strings:\t\t0x%x\n",
		       fdt_off_dt_strings(working_fdt));
		printf("off_mem_rsvmap:\t\t0x%x\n",
		       fdt_off_mem_rsvmap(working_fdt));
		printf("version:\t\t%d\n", version);
		printf("last_comp_version:\t%d\n",
		       fdt_last_comp_version(working_fdt));
		if (version >= 2)
			printf("boot_cpuid_phys:\t0x%x\n",
				fdt_boot_cpuid_phys(working_fdt));
		if (version >= 3)
			printf("size_dt_strings:\t0x%x\n",
				fdt_size_dt_strings(working_fdt));
		if (version >= 17)
			printf("size_dt_struct:\t\t0x%x\n",
				fdt_size_dt_struct(working_fdt));
		printf("number mem_rsv:\t\t0x%x\n",
		       fdt_num_mem_rsv(working_fdt));
		printf("\n");

	/*
	 * Set boot cpu id
	 */
	} else if (strncmp(argv[1], "boo", 3) == 0) {
		unsigned long tmp = simple_strtoul(argv[2], NULL, 16);
		fdt_set_boot_cpuid_phys(working_fdt, tmp);

	/*
	 * memory command
	 */
	} else if (strncmp(argv[1], "me", 2) == 0) {
		uint64_t addr, size;
		int err;
		addr = simple_strtoull(argv[2], NULL, 16);
		size = simple_strtoull(argv[3], NULL, 16);
		err = fdt_fixup_memory(working_fdt, addr, size);
		if (err < 0)
			return err;

	/*
	 * mem reserve commands
	 */
	} else if (strncmp(argv[1], "rs", 2) == 0) {
		if (argv[2][0] == 'p') {
			uint64_t addr, size;
			int total = fdt_num_mem_rsv(working_fdt);
			int j, err;
			printf("index\t\t   start\t\t    size\n");
			printf("-------------------------------"
				"-----------------\n");
			for (j = 0; j < total; j++) {
				err = fdt_get_mem_rsv(working_fdt, j, &addr, &size);
				if (err < 0) {
					printf("libfdt fdt_get_mem_rsv():  %s\n",
							fdt_strerror(err));
					return err;
				}
				printf("    %x\t%08x%08x\t%08x%08x\n", j,
					(u32)(addr >> 32),
					(u32)(addr & 0xffffffff),
					(u32)(size >> 32),
					(u32)(size & 0xffffffff));
			}
		} else if (argv[2][0] == 'a') {
			uint64_t addr, size;
			int err;
			addr = simple_strtoull(argv[3], NULL, 16);
			size = simple_strtoull(argv[4], NULL, 16);
			err = fdt_add_mem_rsv(working_fdt, addr, size);

			if (err < 0) {
				printf("libfdt fdt_add_mem_rsv():  %s\n",
					fdt_strerror(err));
				return err;
			}
		} else if (argv[2][0] == 'd') {
			unsigned long idx = simple_strtoul(argv[3], NULL, 16);
			int err = fdt_del_mem_rsv(working_fdt, idx);

			if (err < 0) {
				printf("libfdt fdt_del_mem_rsv():  %s\n",
					fdt_strerror(err));
				return err;
			}
		} else {
			/* Unrecognized command */
			return CMD_RET_USAGE;
		}
	}
#ifdef CONFIG_OF_BOARD_SETUP
	/* Call the board-specific fixup routine */
	else if (strncmp(argv[1], "boa", 3) == 0)
		ft_board_setup(working_fdt, gd->bd);
#endif
	/* Create a chosen node */
	else if (argv[1][0] == 'c') {
		unsigned long initrd_start = 0, initrd_end = 0;

		if ((argc != 2) && (argc != 4))
			return CMD_RET_USAGE;

		if (argc == 4) {
			initrd_start = simple_strtoul(argv[2], NULL, 16);
			initrd_end = simple_strtoul(argv[3], NULL, 16);
		}

		fdt_chosen(working_fdt, 1);
		fdt_initrd(working_fdt, initrd_start, initrd_end, 1);
	}
	/* resize the fdt */
	else if (strncmp(argv[1], "re", 2) == 0) {
		fdt_resize_action(working_fdt);
	}
	else {
		/* Unrecognized command */
		return CMD_RET_USAGE;
	}

	return 0;
}

/****************************************************************************/

static int fdt_valid(void)
{
	int  err;

	if (working_fdt == NULL) {
		printf ("The address of the fdt is invalid (NULL).\n");
		return 0;
	}

	err = fdt_check_header(working_fdt);
	if (err == 0)
		return 1;	/* valid */

	if (err < 0) {
		printf("libfdt fdt_check_header(): %s", fdt_strerror(err));
		/*
		 * Be more informative on bad version.
		 */
		if (err == -FDT_ERR_BADVERSION) {
			if (fdt_version(working_fdt) <
			    FDT_FIRST_SUPPORTED_VERSION) {
				printf (" - too old, fdt %d < %d",
					fdt_version(working_fdt),
					FDT_FIRST_SUPPORTED_VERSION);
				working_fdt = NULL;
			}
			if (fdt_last_comp_version(working_fdt) >
			    FDT_LAST_SUPPORTED_VERSION) {
				printf (" - too new, fdt %d > %d",
					fdt_version(working_fdt),
					FDT_LAST_SUPPORTED_VERSION);
				working_fdt = NULL;
			}
			return 0;
		}
		printf("\n");
		return 0;
	}
	return 1;
}

/****************************************************************************/

/*
 * Parse the user's input, partially heuristic.  Valid formats:
 * <0x00112233 4 05>	- an array of cells.  Numbers follow standard
 *			C conventions.
 * [00 11 22 .. nn] - byte stream
 * "string"	- If the the value doesn't start with "<" or "[", it is
 *			treated as a string.  Note that the quotes are
 *			stripped by the parser before we get the string.
 * newval: An array of strings containing the new property as specified
 *	on the command line
 * count: The number of strings in the array
 * data: A bytestream to be placed in the property
 * len: The length of the resulting bytestream
 */
static int fdt_parse_prop(char * const *newval, int count, char *data, int *len)
{
	char *cp;		/* temporary char pointer */
	char *newp;		/* temporary newval char pointer */
	unsigned long tmp;	/* holds converted values */
	int stridx = 0;

	*len = 0;
	newp = newval[0];

	/* An array of cells */
	if (*newp == '<') {
		newp++;
		while ((*newp != '>') && (stridx < count)) {
			/*
			 * Keep searching until we find that last ">"
			 * That way users don't have to escape the spaces
			 */
			if (*newp == '\0') {
				newp = newval[++stridx];
				continue;
			}

			cp = newp;
			tmp = simple_strtoul(cp, &newp, 0);
			*(uint32_t *)data = __cpu_to_be32(tmp);
			data  += 4;
			*len += 4;

			/* If the ptr didn't advance, something went wrong */
			if ((newp - cp) <= 0) {
				printf("Sorry, I could not convert \"%s\"\n",
					cp);
				return 1;
			}

			while (*newp == ' ')
				newp++;
		}

		if (*newp != '>') {
			printf("Unexpected character '%c'\n", *newp);
			return 1;
		}
	} else if (*newp == '[') {
		/*
		 * Byte stream.  Convert the values.
		 */
		newp++;
		while ((stridx < count) && (*newp != ']')) {
			while (*newp == ' ')
				newp++;
			if (*newp == '\0') {
				newp = newval[++stridx];
				continue;
			}
			if (!isxdigit(*newp))
				break;
			tmp = simple_strtoul(newp, &newp, 16);
			*data++ = tmp & 0xFF;
			*len    = *len + 1;
		}
		if (*newp != ']') {
			printf("Unexpected character '%c'\n", *newp);
			return 1;
		}
	} else {
		/*
		 * Assume it is one or more strings.  Copy it into our
		 * data area for convenience (including the
		 * terminating '\0's).
		 */
		while (stridx < count) {
			size_t length = strlen(newp) + 1;
			strcpy(data, newp);
			data += length;
			*len += length;
			newp = newval[++stridx];
		}
	}
	return 0;
}

/****************************************************************************/

/*
 * Heuristic to guess if this is a string or concatenated strings.
 */

static int is_printable_string(const void *data, int len)
{
	const char *s = data;

	/* zero length is not */
	if (len == 0)
		return 0;

	/* must terminate with zero */
	if (s[len - 1] != '\0')
		return 0;

	/* printable or a null byte (concatenated strings) */
	while (((*s == '\0') || isprint(*s)) && (len > 0)) {
		/*
		 * If we see a null, there are three possibilities:
		 * 1) If len == 1, it is the end of the string, printable
		 * 2) Next character also a null, not printable.
		 * 3) Next character not a null, continue to check.
		 */
		if (s[0] == '\0') {
			if (len == 1)
				return 1;
			if (s[1] == '\0')
				return 0;
		}
		s++;
		len--;
	}

	/* Not the null termination, or not done yet: not printable */
	if (*s != '\0' || (len != 0))
		return 0;

	return 1;
}


/*
 * Print the property in the best format, a heuristic guess.  Print as
 * a string, concatenated strings, a byte, word, double word, or (if all
 * else fails) it is printed as a stream of bytes.
 */
static void print_data(const void *data, int len)
{
	int j;

	/* no data, don't print */
	if (len == 0)
		return;

	/*
	 * It is a string, but it may have multiple strings (embedded '\0's).
	 */
	if (is_printable_string(data, len)) {
		puts("\"");
		j = 0;
		while (j < len) {
			if (j > 0)
				puts("\", \"");
			puts(data);
			j    += strlen(data) + 1;
			data += strlen(data) + 1;
		}
		puts("\"");
		return;
	}

	if ((len %4) == 0) {
		const u32 *p;

		printf("<");
		for (j = 0, p = data; j < len/4; j ++)
			printf("0x%x%s", fdt32_to_cpu(p[j]), j < (len/4 - 1) ? " " : "");
		printf(">");
	} else { /* anything else... hexdump */
		const u8 *s;

		printf("[");
		for (j = 0, s = data; j < len; j++)
			printf("%02x%s", s[j], j < len - 1 ? " " : "");
		printf("]");
	}
}

/****************************************************************************/

/*
 * Recursively print (a portion of) the working_fdt.  The depth parameter
 * determines how deeply nested the fdt is printed.
 */
static int fdt_print(const char *pathp, char *prop, int depth)
{
	static char tabs[MAX_LEVEL+1] =
		"\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t"
		"\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t";
	const void *nodep;	/* property node pointer */
	int  nodeoffset;	/* node offset from libfdt */
	int  nextoffset;	/* next node offset from libfdt */
	uint32_t tag;		/* tag */
	int  len;		/* length of the property */
	int  level = 0;		/* keep track of nesting level */
	const struct fdt_property *fdt_prop;

	nodeoffset = fdt_path_offset (working_fdt, pathp);
	if (nodeoffset < 0) {
		/*
		 * Not found or something else bad happened.
		 */
		printf ("libfdt fdt_path_offset() returned %s\n",
			fdt_strerror(nodeoffset));
		return 1;
	}
	/*
	 * The user passed in a property as well as node path.
	 * Print only the given property and then return.
	 */
	if (prop) {
		nodep = fdt_getprop (working_fdt, nodeoffset, prop, &len);
		if (len == 0) {
			/* no property value */
			printf("%s %s\n", pathp, prop);
			return 0;
		} else if (len > 0) {
			printf("%s = ", prop);
			print_data (nodep, len);
			printf("\n");
			return 0;
		} else {
			printf ("libfdt fdt_getprop(): %s\n",
				fdt_strerror(len));
			return 1;
		}
	}

	/*
	 * The user passed in a node path and no property,
	 * print the node and all subnodes.
	 */
	while(level >= 0) {
		tag = fdt_next_tag(working_fdt, nodeoffset, &nextoffset);
		switch(tag) {
		case FDT_BEGIN_NODE:
			pathp = fdt_get_name(working_fdt, nodeoffset, NULL);
			if (level <= depth) {
				if (pathp == NULL)
					pathp = "/* NULL pointer error */";
				if (*pathp == '\0')
					pathp = "/";	/* root is nameless */
				printf("%s%s {\n",
					&tabs[MAX_LEVEL - level], pathp);
			}
			level++;
			if (level >= MAX_LEVEL) {
				printf("Nested too deep, aborting.\n");
				return 1;
			}
			break;
		case FDT_END_NODE:
			level--;
			if (level <= depth)
				printf("%s};\n", &tabs[MAX_LEVEL - level]);
			if (level == 0) {
				level = -1;		/* exit the loop */
			}
			break;
		case FDT_PROP:
			fdt_prop = fdt_offset_ptr(working_fdt, nodeoffset,
					sizeof(*fdt_prop));
			pathp    = fdt_string(working_fdt,
					fdt32_to_cpu(fdt_prop->nameoff));
			len      = fdt32_to_cpu(fdt_prop->len);
			nodep    = fdt_prop->data;
			if (len < 0) {
				printf ("libfdt fdt_getprop(): %s\n",
					fdt_strerror(len));
				return 1;
			} else if (len == 0) {
				/* the property has no value */
				if (level <= depth)
					printf("%s%s;\n",
						&tabs[MAX_LEVEL - level],
						pathp);
			} else {
				if (level <= depth) {
					printf("%s%s = ",
						&tabs[MAX_LEVEL - level],
						pathp);
					print_data (nodep, len);
					printf(";\n");
				}
			}
			break;
		case FDT_NOP:
			printf("%s/* NOP */\n", &tabs[MAX_LEVEL - level]);
			break;
		case FDT_END:
			return 1;
		default:
			if (level <= depth)
				printf("Unknown tag 0x%08X\n", tag);
			return 1;
		}
		nodeoffset = nextoffset;
	}
	return 0;
}

/********************************************************************/

U_BOOT_CMD(
	fdt,	255,	0,	do_fdt,
	"flattened device tree utility commands",
	    "addr   <addr> [<length>]        - Set the fdt location to <addr>\n"
#ifdef CONFIG_OF_BOARD_SETUP
	"fdt boardsetup                      - Do board-specific set up\n"
#endif
	"fdt move   <fdt> <newaddr> <length> - Copy the fdt to <addr> and make it active\n"
	"fdt resize                          - Resize fdt to size + padding to 4k addr\n"
	"fdt print  <path> [<prop>]          - Recursive print starting at <path>\n"
	"fdt list   <path> [<prop>]          - Print one level starting at <path>\n"
	"fdt set    <path> <prop> [<val>]    - Set <property> [to <val>]\n"
	"fdt mknode <path> <node>            - Create a new node after <path>\n"
	"fdt rm     <path> [<prop>]          - Delete the node or <property>\n"
	"fdt header                          - Display header info\n"
	"fdt bootcpu <id>                    - Set boot cpuid\n"
	"fdt memory <addr> <size>            - Add/Update memory node\n"
	"fdt rsvmem print                    - Show current mem reserves\n"
	"fdt rsvmem add <addr> <size>        - Add a mem reserve\n"
	"fdt rsvmem delete <index>           - Delete a mem reserves\n"
	"fdt chosen [<start> <end>]          - Add/update the /chosen branch in the tree\n"
	"                                        <start>/<end> - initrd start/end addr\n"
	"NOTE: Dereference aliases by omiting the leading '/', "
		"e.g. fdt print ethernet0."
);

#ifdef CONFIG_OF_LIBFDT
extern int sscanf(const char *buf, const char *fmt, ...);
static uint32_t megre_seq[MAX_MEGRE_NUM];
extern unsigned int KernelImagePosition;
extern BOOT_MODE boot_mode;
char* __find_next_str(char* str)
{
        char* end;
        if(str == NULL)
                return NULL;

        end = str + strlen(str);

        while(str < end)
        {
                if (isalnum(*str)) // is alphabet or number
                        return str;
                str++;
        }

        return NULL;
}

int get_arguments(char* str, int argc, char** argv)
{
        int ret = 0;
        char* sep;

        while ((str = __find_next_str(str)) && ret < argc)
        {
                argv[ret++] = str;
                sep = strchr(str, ','); // devided by

                if (sep == NULL)
                        break;

                *sep = '\0'; // changed to null terminated
                str = sep + 1;

        }

        return ret;
}

static int bootparam_megre_index (char *str)
{
        char* argv[MAX_MEGRE_NUM];
        char tmp_str[128];
        int argc;
        int i;

        if(str == NULL)
        {
                return 0;
        }

        memset(megre_seq, 0 ,MAX_MEGRE_NUM);
        memset(tmp_str, 0 ,128);

        memcpy(tmp_str, str ,strlen(str));
        //.step split param
        argc = get_arguments(tmp_str, MAX_MEGRE_NUM, argv);
        //.try parse param as string label
        for(i = 0; i < argc; ++i)
        {
                sscanf(argv[i], "%d", &megre_seq[i]);
        }

        return argc;
}

static int node_offset(void *fdt, const char *node_path)
{
	int offset = fdt_path_offset(fdt, node_path);
	if (offset == -FDT_ERR_NOTFOUND)
		offset = fdt_add_subnode(fdt, 0, node_path);
	return offset;
}

static int setprop(void *fdt, const char *node_path, const char *property,
		   uint32_t *val_array, int size)
{
	int offset = node_offset(fdt, node_path);
	if (offset < 0)
		return offset;
	return fdt_setprop(fdt, offset, property, val_array, size);
}

static int setprop_string(void *fdt, const char *node_path,
			  const char *property, const char *string)
{
	int offset = node_offset(fdt, node_path);
	if (offset < 0)
		return offset;
	return fdt_setprop_string(fdt, offset, property, string);
}

static int setprop_cell(void *fdt, const char *node_path,
			const char *property, uint32_t val)
{
	int offset = node_offset(fdt, node_path);
	if (offset < 0)
		return offset;
	return fdt_setprop_cell(fdt, offset, property, val);
}

static const void *getprop(const void *fdt, const char *node_path,
			   const char *property, int *len)
{
	int offset = fdt_path_offset(fdt, node_path);

	if (offset == -FDT_ERR_NOTFOUND)
		return NULL;

	return fdt_getprop(fdt, offset, property, len);
}

static uint32_t get_cell_size(const void *fdt)
{
	int len;
	uint32_t cell_size = 1;
	const uint32_t *size_len =  getprop(fdt, "/", "#size-cells", &len);

	if (size_len)
		cell_size = fdt32_to_cpu(*size_len);
	return cell_size;
}

static void merge_fdt_bootargs(void *fdt, const char *fdt_cmdline)
{
	char cmdline[COMMAND_LINE_SIZE];
	const char *fdt_bootargs;
	char *ptr = cmdline;
	int len = 0;

	/* copy the fdt command line into the buffer */
	fdt_bootargs = getprop(fdt, "/chosen", "bootargs", &len);
	if (fdt_bootargs)
		if (len < COMMAND_LINE_SIZE) {
			memcpy(ptr, fdt_bootargs, len);
			/* len is the length of the string
			 * including the NULL terminator */
			ptr += len - 1;
		}

	/* and append the BOOT_CMDLINE */
	if (fdt_cmdline) {
		len = strlen(fdt_cmdline);
		if (ptr - cmdline + len + 2 < COMMAND_LINE_SIZE) {
			*ptr++ = ' ';
			memcpy(ptr, fdt_cmdline, len);
			ptr += len;
		}
	}
	*ptr = '\0';

	setprop_string(fdt, "/chosen", "bootargs", cmdline);
}

/*
 * Convert and fold provided boot cmd into the provided FDT.
 *
 * REturn values:
 *    = 0 -> pretend success
 *    = 1 -> bad boot cmd(may retry with another possible boot cmd pointer)
 *    < 0 -> error from libfdt
 */
int boot_cmd_to_fdt(bootm_headers_t *images, void *fdt, int total_space)
{
	/* In the case of 64 bits memory size, need to reserve 2 cells for
	 * address and size for each bank */
	uint32_t mem_reg_property[2 * 2 * 8];
	int memcount = 0;
	int ret, memsize, i;

	/* let's give it all the room it could need */
	ret = fdt_open_into(fdt, fdt, total_space);
	if (ret < 0)
		return ret;

	/* Append the command line to the device tree*/
	if (do_extend_cmdline)
		merge_fdt_bootargs(fdt, (char *)images->cmdline_start);
	else
		setprop_string(fdt, "/chosen", "bootargs", (char *)images->cmdline_start);

        /*Append mem reg to the device tree*/
	for (i = 0; i < CONFIG_NR_DRAM_BANKS; i++) {
        	if (memcount >= sizeof(mem_reg_property)/4)
        		continue;
        	if (!images->kbd->bi_dram[i].size)
        		continue;
        	memsize = get_cell_size(fdt);

        	if (memsize == 2) {
        		/* if memsize is 2, that means that
        		 * each data needs 2 cells of 32 bits,
        		 * so the data are 64 bits */
        		uint64_t *mem_reg_prop64 =
        			(uint64_t *)mem_reg_property;
        		mem_reg_prop64[memcount++] =
        			cpu_to_fdt64(images->kbd->bi_dram[i].start);
        		mem_reg_prop64[memcount++] =
        			cpu_to_fdt64(images->kbd->bi_dram[i].size);
        	} else {
        		mem_reg_property[memcount++] =
        			cpu_to_fdt32(images->kbd->bi_dram[i].start);
        		mem_reg_property[memcount++] =
        			cpu_to_fdt32(images->kbd->bi_dram[i].size);
        	}
	}

	if (memcount)
		setprop(fdt, "/memory", "reg", mem_reg_property, 4 * memcount * memsize);

        /*Append initrd to the device tree*/
        if(0)
        {
        	setprop_cell(fdt, "/chosen", "linux,initrd-start",(uint32_t)images->rd_start);
        	setprop_cell(fdt, "/chosen", "linux,initrd-end",(uint32_t)images->rd_end);
        }

	return fdt_pack(fdt);
}

/*
 * Megre DTBO fdt into default FDT.
 *
 *parameter:
 *base_fdt_addr:   default FDT DATA DDR addr,must fdt format
 *base_fdt_size:    default FDT DATA size
 *dtbo_fdt_addr:    DTBO fdt DATA DDR addr,must fdt format
 *dtbo_fdt_size:     DTBO fdt DATA size
 *
 * REturn values:
 *    = 0 -> pretend success
 *    = -1 -> megre dtb/dtbo failed
 *    = -2 -> input parameter invalid
 */

int Fdt_Megre(void *base_fdt_addr, unsigned int base_fdt_size, void *dtbo_fdt_addr, unsigned int dtbo_fdt_size)
{
    struct fdt_header * tmp_megred_fdt_addr = NULL;

    if((base_fdt_addr == NULL)||(base_fdt_size == 0))
    {
        printf("%s input dtb base parameter invalid\n",__FUNCTION__);
        hang();
        return -2;
    }

    if((dtbo_fdt_addr == NULL)||(dtbo_fdt_size == 0))
    {
        printf("%s can not megre dtbo,because input dtbo data invalid.if you want magre DTBO into DTB,please double check DTBO data\n",__FUNCTION__);
        return -2;
    }

    tmp_megred_fdt_addr = ufdt_apply_overlay(base_fdt_addr, base_fdt_size, dtbo_fdt_addr , dtbo_fdt_size);

    if(tmp_megred_fdt_addr == NULL){
        printf("%s megre dtb/dtbo failed\n",__FUNCTION__);
        return -1;
    }else{
        memcpy((void *)CONFIG_DTB_MEGRE_BUFFER_ADDR,tmp_megred_fdt_addr, fdt_totalsize(tmp_megred_fdt_addr));
        ufdt_overlay_malloc_buffer_free();/*this free buffer is be malloced in function <ufdt_apply_overlay>*/
    }

    return 0;
}

/*
 * insert boot cmd into DTB.
 *
 *parameter:
 *boot_cmd:       boot cmd pointer
 *fdt:                  inserte DTB pointer,must fdt format
 *total_space:     DTBO fdt DATA DDR addr,must fdt format
 *
 * REturn values:
 *    = 0 -> insert success
 *    = -1 -> insert dtb failed
 */

int Fdt_Insert_boot_cmd_to_DTB(bootm_headers_t *images, void *fdt, int total_space)
{
    if((images == NULL)||(fdt == NULL)||(total_space == 0))
    {
        printf("%s input parameter invalid\n",__FUNCTION__);
        return -1;
    }

    if(total_space > 0x4000)
    {
        printf("%s DTB size is too long(size=0x%x)\n",__FUNCTION__,total_space);
        return -1;
    }

    if(boot_cmd_to_fdt(images, fdt, total_space) != 0)
    {
        printf("%s insert boot cmd into FDT failed\n",__FUNCTION__);
        return -1;
    }

    return 0;
}

/*
 * find and match DTB DATA.
 *
 *parameter:
 *base_fdt_addr: DTB base img pointer,DATA must fdt format
 *base_img_size: DTB base img size
 *base_fdt_addr: return match DTB base pointer
 *base_fdt_size:return match DTB base DDR size
 *
 *REturn values:
 *    = 0 -> find and match DTB success
 *    = -1 -> find and match DTB failed
 */

int Fdt_Find_Match_DTB_DATA(void *base_img_addr, unsigned long base_img_size, unsigned int *base_fdt_addr, unsigned int *base_fdt_size)
{
        struct fdt_header * fdt_img_header = NULL;
        int  nodeoffset;	/* node offset from libfdt */
        const void *nodep;	/* property node pointer */
        int len = 0;/* length of the property */
        char* vendor_soc_str[2];
        char vendor_soc_str_tmp[128];

        if((base_img_addr == NULL)||(base_img_size == 0))
        {
            printf("%s input DTB base addr invalid\n",__FUNCTION__);
            hang();
            return -1;
        }

        if((base_fdt_addr == NULL)||(base_fdt_size == NULL))
        {
            printf("%s input DTB parameter invalid\n",__FUNCTION__);
            return -1;
        }

        fdt_img_header = (struct fdt_header *)base_img_addr;
        memset(vendor_soc_str, 0, sizeof(char*)*2);
        memset(vendor_soc_str_tmp, 0, 128);
        *base_fdt_addr = 0;
        *base_fdt_size = 0;

        while(fdt_img_header ->magic == fdt32_to_cpu(FDT_MAGIC))
        {
            /*DTB base find compatible name*/
            nodeoffset = fdt_path_offset ((void *)fdt_img_header, "/");
            if(nodeoffset >= 0 )
            {
                nodep = fdt_getprop ((void *)fdt_img_header, nodeoffset, "compatible", &len);
                if (len == 0) {
                	printf("%s %s no vaule\n", "/", "compatible");
                } else if (len > 0) {
                    memcpy(vendor_soc_str_tmp,(char*)nodep, len);
                    get_arguments(vendor_soc_str_tmp, 2, vendor_soc_str);
                    /*match source code*/
                    if(strncmp(vendor_soc_str[0], CONFIG_SYS_SOC, sizeof(CONFIG_SYS_SOC)) == 0)
                    {
                         if(strncmp(vendor_soc_str[1],Config_DTE_Bootcode_HW_Value,sizeof(Config_DTE_Bootcode_HW_Value)) == 0)
                         {
                                printf("DTB %s %s is match \n", "vendor", "SOC");
                                nodeoffset = fdt_path_offset ((void *)fdt_img_header, "/firmware/android");
                                nodep = fdt_getprop ((void *)fdt_img_header, nodeoffset, "hardware", &len);
                                if (len > 0) {
                                    if(strncmp(nodep,Config_DTE_Bootcode_HW_Value,len) == 0)
                                    {
                                        printf("DTB %s %s is match \n", "hardware", Config_DTE_Bootcode_HW_Value);
                                        *base_fdt_addr = (unsigned int)fdt_img_header;
                                        *base_fdt_size = cpu_to_fdt32(fdt_img_header->totalsize);
                                        break;
                                    }
                                } else {
                                     *base_fdt_addr = (unsigned int)fdt_img_header;
                                     *base_fdt_size = cpu_to_fdt32(fdt_img_header->totalsize);
                                     //printf("%s no vaule\n", "/firmware/hardware");
                                     break;
                                }
                          }
                     }
                } else {
                	printf ("%s(%d): %s\n",__FUNCTION__, __LINE__, fdt_strerror(len));
                }
            }

            fdt_img_header += cpu_to_fdt32(fdt_img_header ->totalsize);
        };

        //printf("*base_fdt_addr = 0x%x\n",*base_fdt_addr);
        //printf("*base_fdt_size = 0x%x\n",*base_fdt_size);

        if((*base_fdt_addr == 0)||(*base_fdt_size == 0))
        {
            printf("%s can not find matched DTB base\n",__FUNCTION__);
            hang();
            return -1;
        }

        return 0;
}

/*
 * find and match DTBO DATA.
 *
 *parameter:
 *dtbo_img_addr:DTBO img DATA pointer
 *dtbo_img_size:DTBO img size
 *dtbo_megre_num: return need megre DTBO totle num
 *dtbo_fdt_addr_list:return match DTBO DDR addr list
 *dtbo_fdt_size_list:return match DTBO DATA size
 *
 *REturn values:
 *    = 0 -> find and match DTBO success
 *    = -1 -> find and match DTBO failed
 */

int Fdt_Find_Match_DTBO_DATA(void *dtbo_img_addr, unsigned long dtbo_img_size, unsigned int * dtbo_megre_num, unsigned int* dtbo_fdt_addr_list, unsigned int* dtbo_fdt_size_list)
{
        struct dt_table_header * dtbo_img_header = NULL;
        struct dt_table_entry * dtbo_entry = NULL;
        unsigned int  dtbo_entry_num = 0;
        char *commandline = getenv("bootargs");
        char *megre_index_sequence = NULL;
        int megre_total_num = 0;
        int i = 0;
        uint32_t custom_id = 0;
        char dtbo_idx_str[512];

        if((dtbo_img_addr == NULL )||(dtbo_img_size == 0) \
            ||(dtbo_fdt_addr_list == NULL)||(dtbo_fdt_size_list == NULL ))
        {
            printf("%s input DTBO parameter invalid\n",__FUNCTION__);
            return -1;
        }

        memset(dtbo_idx_str, 0, 512);
        /*get cmdline megre index sequence*/
        megre_index_sequence = strstr(commandline,"androidboot.dtbo_idx");
        if(megre_index_sequence == NULL)
        {
            printf("%s not find androidboot.dtbo_idx in cmdline,please confirm only one dtbo to insert dtb base.\n",__FUNCTION__);
             megre_total_num = 1;
        }else{
             sscanf(megre_index_sequence+strlen("androidboot.dtbo_idx="), "%s" , dtbo_idx_str);
             megre_total_num = bootparam_megre_index(dtbo_idx_str);
             if(megre_total_num <= 0)
             {
                    printf("%s not find dtbo megre index sequence in cmdline,please confirm only one dtbo to insert dtb base.\n",__FUNCTION__);
                    megre_total_num = 1;
             }
        }

        *dtbo_megre_num = megre_total_num;

        /*Get custom str*/
        sscanf(CONFIG_SYS_BOARD, "tv%d", &custom_id);

        /*search and match dtbo*/
        dtbo_img_header = (struct dt_table_header *)dtbo_img_addr;
        if(cpu_to_fdt32(dtbo_img_header->magic) == DT_TABLE_MAGIC)
        {
            dtbo_entry = (struct dt_table_entry * )(dtbo_img_addr + cpu_to_fdt32(dtbo_img_header->header_size));

            if(megre_total_num == 1)/*only one dtbo need to insert*/
            {
                for(dtbo_entry_num=0; dtbo_entry_num < cpu_to_fdt32(dtbo_img_header->dt_entry_count);dtbo_entry_num++)
                {
                      if(cpu_to_fdt32(dtbo_entry->custom[0]) == custom_id)
                      {
                            printf("DTBO %s[%03d]  match \n", "custom ID",custom_id);
                            *dtbo_fdt_addr_list=(unsigned int)(dtbo_img_addr+cpu_to_fdt32(dtbo_entry->dt_offset));
                            *dtbo_fdt_size_list=cpu_to_fdt32(dtbo_entry->dt_size);
                            break;
                      }else{
                            printf("DTBO %s[%03d]  not find \n", "custom ID",cpu_to_fdt32(dtbo_entry->custom[0]) );
                            return -1;
                      }

                      dtbo_entry += cpu_to_fdt32(dtbo_img_header->dt_entry_size);
                }
            }else{/*multi-dtbo need to insert*/
                for( i = 0; i < megre_total_num; i++)
                {
                    for(dtbo_entry_num=0; dtbo_entry_num < cpu_to_fdt32(dtbo_img_header->dt_entry_count);dtbo_entry_num++)
                    {
                          if(cpu_to_fdt32(dtbo_entry->custom[0]) == custom_id)
                          {
                                if(cpu_to_fdt32(dtbo_entry->id) == megre_seq[i])
                                {
                                    *(dtbo_fdt_addr_list+i*sizeof(unsigned int*))=(unsigned int)(dtbo_img_addr+cpu_to_fdt32(dtbo_entry->dt_offset));
                                    *(dtbo_fdt_size_list+i*sizeof(unsigned int*))=cpu_to_fdt32(dtbo_entry->dt_size);
                                    break;
                                }
                          }

                          dtbo_entry += cpu_to_fdt32(dtbo_img_header->dt_entry_size);
                    }
                }
            }
        }else{
            printf("%s DTBO data invalid\n",__FUNCTION__);
            return -1;
        }

        return 0;
}

/*
 * find and match DTBO DATA.
 *
 *parameter:
 *images: bootm_headers_t images parameter
 *
 *REturn values:
 *    = 0 -> fdt prepare success
 *    = -1 ->fdt prepare failed
 */
int Fdt_prep_linux(bootm_headers_t *images)
{
        int ret = 0;
        unsigned int DTB_base_addr = 0;
        unsigned int DTB_base_size = 0;
        unsigned int DTBO_megre_num = 0;
        unsigned int DTBO_addr_list[MAX_MEGRE_NUM];
        unsigned int DTBO_size_list[MAX_MEGRE_NUM];
        char * megre_space = (char *)CONFIG_DTB_MEGRE_BUFFER_ADDR;
        int i = 0;

        memset(DTBO_addr_list, 0,sizeof(unsigned int)*MAX_MEGRE_NUM);
        memset(DTBO_size_list, 0,sizeof(unsigned int)*MAX_MEGRE_NUM);
        ret = Fdt_Find_Match_DTB_DATA(images->ft_addr, images->ft_len, &DTB_base_addr, &DTB_base_size);
        if((ret != 0)||(DTB_base_addr == 0)||(DTB_base_size == 0))
            return -1;
        if(DTB_base_addr != CONFIG_DTB_MEGRE_BUFFER_ADDR)//android dtb at boot partition,kernel only&rescue dtb at CONFIG_DTB_MEGRE_BUFFER_ADDR
        {
            memset(megre_space, 0, CONFIG_DTB_MEGRE_BUFFER_SIZE);
            memcpy(megre_space, (void *)DTB_base_addr, DTB_base_size);
        }

        ///printf("KernelImagePosition=%d boot_mode=%d\n", KernelImagePosition, boot_mode);

        if((KernelImagePosition == 0)&&(boot_mode != BOOT_RESCUE_MODE) )
        {
            ret = Fdt_Find_Match_DTBO_DATA(images->dtbo_fdt_addr, images->dtbo_fdt_size, &DTBO_megre_num, DTBO_addr_list, DTBO_size_list);
            if((ret == 0)&&(DTBO_megre_num > 0)&&(DTBO_megre_num < MAX_MEGRE_NUM))
            {
                for(i = 0; i < DTBO_megre_num;i++)
                {
                    DTB_base_size = fdt_totalsize((struct fdt_header *)megre_space);
                    ret = Fdt_Megre(megre_space, DTB_base_size, (void *)DTBO_addr_list[i], DTBO_size_list[i]);
                    if(ret != 0)
                    {
                        printf("Megre %d DTBO fail!plz check it\n",i);
                        return -1;
                     }
                }
            }
        }

        ret = Fdt_Insert_boot_cmd_to_DTB(images, megre_space, 0x4000);/*0x4000 is the size of gd->bd->bi_boot_params*/
        if(ret != 0)
            return -1;

        memcpy((void*)images->kbd->bi_boot_params, megre_space, 0x4000);

        return 0;
}

/*
 * Kernel_mode_DTB_Verify.
 *
 *parameter:
 *images: bootm_headers_t images parameter
 *
 *REturn values:
 *    = 0 -> Verify Rescue DTB success
 *    = -1 ->Verify Rescue DTB
 */

int Kernel_mode_DTB_Verify(bootm_headers_t *images)
{
        struct fdt_header * fdt_img_header = (struct fdt_header *)CONFIG_DTB_MEGRE_BUFFER_ADDR;

        if(fdt_img_header ->magic == fdt32_to_cpu(FDT_MAGIC))
        {
            images->ft_addr=(char *)(CONFIG_DTB_MEGRE_BUFFER_ADDR);
            images->ft_len=cpu_to_fdt32(fdt_img_header ->totalsize);
            return 0;
        }else{
            if (boot_mode == BOOT_RESCUE_MODE)
                printf("rescue kernel DTB data invalid\n");
            else
                printf("normal kernel DTB data invalid\n");
            return -1;
        }
}

struct file_info
{
	uint file_type;		/*fw type eg: for  0 for linux_kernel*/
	uint file_addr;		/*ddr address*/
	uint file_size;		/*file size*/
};

#ifdef CONFIG_CMD_MMC
#include <malloc.h>
extern int get_tftp_fileinfo(char *file_path,struct file_info *ddr_fileinfo);

int fdt_megre_test (cmd_tbl_t * cmdtp, int flag, int argc, char * const argv[])
{
#ifndef CONFIG_SYS_RTK_SPI_FLASH
        struct file_info basefileinfo = {0};
        struct file_info overlayfileinfo = {0};
        char* basedtb_buffer = NULL;
        int ret = -1;
        char tftpput_cmd[512] = {0};
        struct fdt_header * fdt_img_header = NULL;

	if(argc < 4){
            printf("input param error\n");
	    ret = CMD_RET_FAILURE;
            goto exit;
	}

        basedtb_buffer=(char *)malloc(0x4000);
        if(basedtb_buffer == NULL){
            printf("malloc buffer error\n");
            ret = CMD_RET_FAILURE;
            goto exit;
        }

        memset(basedtb_buffer, 0 ,0x4000);

	if(get_tftp_fileinfo(argv[1],&basefileinfo) != 0)
        {
            printf("get base dtb file fail\n");
            ret = CMD_RET_FAILURE;
            goto exit;
        }

        if(basefileinfo.file_size <= 0x4000){
            memcpy(basedtb_buffer, (void *)basefileinfo.file_addr, basefileinfo.file_size);
        }else{
           printf("base dtb file too large\n");
           ret = CMD_RET_FAILURE;
           goto exit;
        }

    	if(get_tftp_fileinfo(argv[2],&overlayfileinfo) != 0)
        {
            printf("get overlay dtb file fail\n");
            ret = CMD_RET_FAILURE;
            goto exit;
        }

        //printf("basedtb_buffer=0x%x basefileinfo.file_size=%d\n",(uint32_t)basedtb_buffer,basefileinfo.file_size);
        //printf("overlayfileinfo.file_addr=0x%x overlayfileinfo.file_size=%d\n",overlayfileinfo.file_addr,overlayfileinfo.file_size);

        Fdt_Megre(basedtb_buffer, basefileinfo.file_size, (void *)overlayfileinfo.file_addr, overlayfileinfo.file_size);
        fdt_img_header = (struct fdt_header *)CONFIG_DTB_MEGRE_BUFFER_ADDR;

        snprintf(tftpput_cmd, 512, "%s 0x%x 0x%x %s", "tftpput", (uint32_t)CONFIG_DTB_MEGRE_BUFFER_ADDR, cpu_to_fdt32(fdt_img_header ->totalsize), argv[3]);
        run_command_list(tftpput_cmd, -1 , 0);
 exit:
        if(basedtb_buffer != NULL)
            free(basedtb_buffer);
        return ret;
#else
	printf("SPI Solution not support this cmd!\r\n");
	return 0;
#endif
}

U_BOOT_CMD(
	fdt_megre,	255,	0,	fdt_megre_test,
	"fdt megre commands via tftp",
	"<base dtb file> <overlay dtb file> <meger dtb file>"
);
#endif
#endif
