/*
 * Driver for NAND support, Rick Bronson
 * borrowed heavily from:
 * (c) 1999 Machine Vision Holdings, Inc.
 * (c) 1999, 2000 David Woodhouse <dwmw2@infradead.org>
 *
 * Ported 'dynenv' to 'nand env.oob' command
 * (C) 2010 Nanometrics, Inc.
 * 'dynenv' -- Dynamic environment offset in NAND OOB
 * (C) Copyright 2006-2007 OpenMoko, Inc.
 * Added 16-bit nand support
 * (C) 2004 Texas Instruments
 *
 * Copyright 2010, 2012 Freescale Semiconductor
 * The portions of this file whose copyright is held by Freescale and which
 * are not considered a derived work of GPL v2-only code may be distributed
 * and/or modified under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */

#include <common.h>
#include <linux/mtd/mtd.h>
#include <command.h>
#include <watchdog.h>
#include <burn.h>
#include <malloc.h>
#include <asm/byteorder.h>
#include <asm/arch/sys_flash_layout.h>
#include "../drivers/mtd/spi/rtk_spi_nand.h"

extern void printf_spi_nand_info(struct mtd_info *mtd);
extern int spi_nand_read_random(struct mtd_info *mtd,char *buf,
	int start_addr, int length,int raw);
extern  void dump_byte(char *buf, int buf_size);
extern  void dump_sn_reg();
extern void dump_sn_bi_table(struct mtd_info *mtd);
extern  void dump_sn_ibt(struct mtd_info * mtd);
extern  void dump_sn_bbt(struct mtd_info * mtd);
extern  void dump_sn_rbt(struct mtd_info * mtd);;
extern  void dump_sn_log2phy_talbe(struct mtd_info * mtd);
extern  int spi_nand_write_random(struct mtd_info *mtd,char *buf,
	int start_addr, int length,int raw);	
extern  int spi_nand_write_block(struct mtd_info *mtd,int start_page,
	char *buf_content, int  block_tag);
extern int spi_nand_read_page(struct mtd_info *mtd,int start_page,int page_num, char *buf_content);
extern int get_free_rab_num(struct mtd_info *mtd);
extern int sn_get_remap_block(struct mtd_info *mtd,int bad_log_block);
extern int sn_get_free_bbt_block(struct mtd_info *mtd,int bad_log_block);
extern int sn_get_free_block(struct mtd_info *mtd,int bad_log_block);
extern int sn_get_log2phy_block(struct mtd_info *mtd,int log_block);
extern int sfc_read_page_to_cache_b(const unsigned int block_addr,const unsigned int page_addr);
extern void sfc_read_cache_to_mem_cpu(unsigned int block_addr,UINT8 *dst,
	unsigned int offset,unsigned int len);
#ifdef CONFIG_MTD_TABLE_TEST
extern int set_erase_fail;
extern int set_write_fail;
extern int set_read_fail;
#endif

static int read_measure;


static void spi_nand_print_and_set_info(int index)
{
    struct mtd_info *spi_nand = &spi_nand_info[index];
    printf_spi_nand_info(spi_nand);   
}

static void scan_spi_nand_block_status(struct mtd_info  *mtd)
{
    int loop = 0,bad_block_num = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("\n Device  bad blocks status:\n");
    for(loop = 0; loop < this->total_block_num;loop++)
    {
        if(this->is_good_block(mtd,loop)!= 0)
        {
            printf("%d is bad block \n",loop);
            bad_block_num++;
        }
    }
    printf("Total block number %d  bad block_num %d \n",
        this->total_block_num,bad_block_num);
}

static void spi_nand_get_bootcode_status(struct mtd_info *mtd)
{
    int block_id = 0, end_block = 0, loop = 0;
    unsigned char block_status = 0xff;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    end_block = this->bootcode_end_phy_block;
    printf("Bootcode area block status([xx, xx]-->[blocknum, blockstatus]): end block %d \n",
        end_block);
    for (block_id = 0; block_id < end_block; block_id++)
    {
        block_status = mtd->block_status_get(mtd, block_id*this->block_size);
        loop++;
        printf("[0x%04x, 0x%02x]\t", block_id, block_status);
        if(loop%8 == 0)
        {
            printf("\n");
        }
    }
}

static void dump_content(struct mtd_info *mtd,ulong start_addr, ulong length,int raw,
	char *content_buf)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    char *buf = (char *)malloc(sizeof(char)*this->page_size);
    int dump_len = 0;
	printf("%s dump  start_addr %lu size %lu \n",__FUNCTION__,start_addr,length);
	
    if(buf != NULL)
    {
    	  while(length > 0)
    	  {
    	      dump_len = length > this->page_size ? this->page_size:length;
              if(spi_nand_read_random(mtd, buf,start_addr,dump_len,raw) == 0)
              {
                  if(content_buf != NULL)
                  {
                      memcpy(content_buf,buf,dump_len);
                      content_buf = content_buf + dump_len;
                  }
                  else
                  {
                          dump_byte(buf,dump_len);
                  }
                  
                  start_addr  += dump_len;
                  length      -= dump_len;    
              }
              else
              {
        	      printf("%s dump read error start_addr %lu size %lu \n",
        	       __FUNCTION__,start_addr,length);
                  break;
              }
        }
        free(buf);
    }
    else
    {
        printf("%s can't malloc 0x%x size \n",__FUNCTION__,this->page_size);
    }
}

static void dump_by_cpu(struct mtd_info *mtd,ulong page_addr)
{
    unsigned int block_addr = 0;
    struct spi_nand_chip *this=(struct spi_nand_chip *)mtd->priv;
    char *buf = (char *)malloc(sizeof(char)*this->page_size);
	if(read_measure == 0){
    	printf("%s page_addr %d size %u \n",
			 __FUNCTION__,(int)page_addr,this->page_size);
	}
    if(buf != NULL)
    {
        /*read to cache*/
        block_addr = page_addr/this->block_per_page;
        if(sfc_read_page_to_cache_b(block_addr,page_addr) == 0)
        {
            /*read cache to memeroy*/
            sfc_read_cache_to_mem_cpu(block_addr,buf,0,this->page_size);
			if(read_measure == 0){
            	dump_byte(buf,this->page_size);
			}
        }
        else
        {
            printf("%s  read error page_addr %d size %u \n",
			 __FUNCTION__,(int)page_addr,this->page_size);
        }

        free(buf);
    }
}

static void write_oob(struct mtd_info *mtd,ulong page_addr,int oob)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("%s page_addr %d oob 0x%x \n",__FUNCTION__,(int)page_addr,oob); 
    memcpy(&(this->g_oobbuf[this->magic_tag_offset]),(char*)&oob,sizeof(int));
    dump_byte(this->g_oobbuf,this->oob_size);
    
    *(int *)(&this->g_oobbuf[this->magic_tag_offset]) = oob;
    dump_byte(this->g_oobbuf,this->oob_size);
    if(this->write_oob(mtd,page_addr,this->g_oobbuf,this->oob_size)!= 0)
    {
        printf("%s page_addr %d oob 0x%x failed! \n",__FUNCTION__,(int)page_addr,oob); 
    }
}

static void  mark_badblock(struct mtd_info *mtd,ulong block_addr)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("%s mark %d bad block\n",__FUNCTION__,(int)block_addr);
    memset(this->g_oobbuf,0xFF,this->oob_size);
    this->g_oobbuf[this->bad_tag_offset]= 0;
	this->g_oobbuf[this->magic_tag_offset] = 0x33;
    dump_byte(this->g_oobbuf,this->oob_size);
    this->write_oob(mtd,block_addr*this->block_per_page,
		this->g_oobbuf,this->oob_size);
    this->read_oob(mtd,block_addr*this->block_per_page,
		this->g_oobbuf,this->oob_size);
    dump_byte(this->g_oobbuf,this->oob_size);
}

static void set_cache_mode(struct mtd_info *mtd,int cache_mode)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("%s before cache %d set cahce_mode %d\n",__FUNCTION__,this->support_cache_mode,cache_mode);
    if((cache_mode<0)||(cache_mode>2))
    {
        printf("%s set wrong cache mode %d (0:normal 1:dual 2:quad 3:quad_IO)\n",__FUNCTION__,cache_mode);
    }
    else{
        this->support_cache_mode = cache_mode;
    }
}

void dump_page_oob(struct mtd_info *mtd,ulong page_addr)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("%s page_addr %d \n",__FUNCTION__,(int)page_addr);
    if(this->read_oob(mtd,page_addr,this->g_oobbuf,this->oob_size) == 0)
    {
        dump_byte(this->g_oobbuf,this->oob_size);
    }
    else
    {
        printf("%s read page_addr %d error! \n",__FUNCTION__,(int)page_addr);
    }
}

static void erase_spi_nand_block(struct mtd_info *mtd,ulong block_addr,ulong size)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    int block_id = 0;
    for(block_id = 0; block_id < size; block_id++)
    {
        if(this->erase_block(mtd,block_addr+block_id) == 0)
        {
            printf("%s erase block %d successful!\n",__FUNCTION__,(int)(block_addr+block_id));
        }
    }
}

void get_random_content(u_char *content, int content_size)
{
    int loop = 0;
    for(loop = 0; loop < content_size; loop=loop+2)
    {
        *content++ = 0x52;
        *content++ = 0x33;
    }
}

int compare_memory(u_char *src_mem,u_char *dst_mem, int memsize)
{
    int loop = 0, ret = 0;
    for(loop = 0; loop< memsize; loop++)
    {
        if((*src_mem) != (*dst_mem))
        {
            printk("[%s]loop%d Src 0x%x dest 0x%x !!\n", __FUNCTION__,loop,*src_mem,*dst_mem); 
            ret = -1;
        }
        src_mem++;
        dst_mem++;
    }
    return ret;
}

static int phy_block_wp_test(struct mtd_info *mtd,int block_addr, 
	    u_char *block_content,u_char *bak_block_content)
{
    int ret = 0,phy_page = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    phy_page = block_addr*this->block_per_page;
	/*Write patten to nandflash block*/
    if((ret = this->erase_block(mtd, block_addr))!=0)
    {
        printk("[%s]block %d erase block error!!\n", __FUNCTION__,block_addr);
    }
	else{
        if((ret = spi_nand_write_block(mtd,phy_page,block_content,0xff))!= 0)
        {
            printk("[%s]block %d  write block error !!\n", __FUNCTION__,block_addr);
        }
        else{
            if((ret = spi_nand_read_page(mtd,phy_page,this->block_per_page, bak_block_content)) !=0)
            {
                printk("[%s]block %d  read block error !!\n", __FUNCTION__,block_addr);
            }
            else
            {
                ret = compare_memory(block_content,bak_block_content,1<<this->block_shift);
            }
        }
    }
    return ret;
}

static void sn_mul_block_test(struct mtd_info *mtd,int block_addr, int block_num)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    int  write_times = 0,block_id = 0,ret = 0;
    if(block_num > 0)
	{
        u_char *block_content     =  (u_char *)malloc(1<<this->block_shift);
        u_char *bak_block_content =  (u_char *)malloc(1<<this->block_shift);
        printk("[%s]block %d num %d \n", __FUNCTION__,block_addr,block_num);
        if((block_content != NULL)&&(bak_block_content != NULL))
        {
            get_random_content(block_content,1<<this->block_shift);
            while(true)
            {
                for(block_id = 0; block_id < block_num; block_id++)
                {
                    if((ret = phy_block_wp_test(mtd,(block_addr+block_id),block_content,bak_block_content)) == 0)
                    {
		                printk("[%s]block %d write_times %d compare OK !!\n", __FUNCTION__,(block_addr+block_id),write_times);
	                }
                    else
                    {
                        printk("[%s]block %d write_times %d compare error !!\n", __FUNCTION__,(block_addr+block_id),write_times);
	                    break;
                    }
                }
                if(ret == 0)
                {
                    write_times++;
                }
                else
                {
                    break;
                }
             }
		     free(block_content);
             free(bak_block_content);
        }
        else
        {
            printk("[%s]malloc error size 0x%x!!\n", __FUNCTION__,1<<this->block_shift);
        }
    }
}

static void sn_single_block_test(struct mtd_info *mtd,int block_addr)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    int  write_times = 0;
    u_char *block_content     =  (u_char *)malloc(1<<this->block_shift);
    u_char *bak_block_content =  (u_char *)malloc(1<<this->block_shift);
    printk("[%s]block %d size 0x%x!!\n", __FUNCTION__,block_addr,(1<<this->block_shift));
    if((block_content != NULL)&&(bak_block_content != NULL))
    {
        get_random_content(block_content,1<<this->block_shift);
        while(true)
        {
            /*Write patten to nandflash block*/           
            if(phy_block_wp_test(mtd,block_addr,block_content,bak_block_content) == 0)
		    {
		        printk("[%s]block %d write_times %d compare OK !!\n",__FUNCTION__,block_addr,write_times);
		    }
		    else
		    {
		        printk("[%s]block %d write_times %d compare error !!\n",__FUNCTION__,block_addr,write_times);
		        break;
		    }
            write_times++;
         }
		 free(block_content);
         free(bak_block_content);
    }
	else
	{
		printk("[%s]malloc error size 0x%x!!\n", __FUNCTION__,1<<this->block_shift);
	}
}

static void wp_block_test(struct mtd_info *mtd,int block_addr,int block_num)
{
    sn_mul_block_test(mtd,block_addr,block_num);
    sn_single_block_test(mtd,(block_addr+block_num));
}

static void table_wp_test(struct mtd_info *mtd,int block_addr)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    int wp_times = 0;
    size_t retlen = 0;
    uint64_t addr = block_addr*this->block_size;
    struct erase_info instr = {0};
    u_char *block_content     =  (u_char *)malloc(1<<this->block_shift);
    u_char *bak_block_content =  (u_char *)malloc(1<<this->block_shift);
	instr.mtd  = mtd;
	instr.addr = block_addr*this->block_size;
	instr.len  = this->block_size;
    if((block_content != NULL)&&(bak_block_content != NULL))
    {
        get_random_content(block_content,1<<this->block_shift);
        while(true)
        {
            if(mtd->erase(mtd,&instr) != 0)
            {
                printf("%s erase failed time %d \n",__FUNCTION__,wp_times);
                break;
            }
            if(mtd->write(mtd,addr,this->block_size,&retlen,block_content)!= 0)
            {
                printk("[%s]block %d write_times %d writeblock error !!\n", __FUNCTION__,block_addr,wp_times);
                break;
            }
			  
            if(mtd->read(mtd,addr,this->block_size,&retlen,bak_block_content)!=0)
			{
                printk("[%s]block %d write_times %d readblock error !!\n", __FUNCTION__,block_addr,wp_times);
                break;
            }
					  
            if(compare_memory(block_content,bak_block_content,1<<this->block_shift) == 0)
            {
                printk("[%s]block %d write_times %d compare OK !!\n", __FUNCTION__,block_addr,wp_times);
            }
            else
            {
                printk("[%s]block %d write_times %d compare error !!\n", __FUNCTION__,block_addr,wp_times);
                break;
            }
            wp_times++;
        }
        free(block_content);
        free(bak_block_content);
    }
    else
    {
        printk("[%s]malloc error size 0x%x!!\n", __FUNCTION__,1<<this->block_shift);
    }
}

static void sn_bbt_collect_test(struct mtd_info *mtd,int block_addr)
{
    int remap_block = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("%s block_addr %d\n",__FUNCTION__,block_addr);
    /*get free current rba number -->use the same logical block*/
    while(this->cur_rba_num > 0)
    {
        /*reduce rba number in the memory*/
        remap_block  = sn_get_remap_block(mtd, block_addr);
        printf("%s reduce remap_block is %d cur rab num %d \n",
			__FUNCTION__,remap_block,this->cur_rba_num);
    }
    dump_sn_bbt(mtd);
	printf("%s log block %d phyblock %d \n",__FUNCTION__,
		block_addr,sn_get_log2phy_block(mtd,block_addr));
	
	/*start bbt collect*/
	printf("%s start bbt collect mode \n",__FUNCTION__);
    remap_block = sn_get_free_bbt_block(mtd,block_addr);
    printf("%s remap_block is %d cur rab num %d \n",
			__FUNCTION__,remap_block,this->cur_rba_num);
    dump_sn_bbt(mtd);
	printf("%s log block %d phyblock %d \n",__FUNCTION__,
		block_addr,sn_get_log2phy_block(mtd,block_addr));
}

static void sn_rbt_collect_test(struct mtd_info *mtd,int block_addr)
{
    int remap_block = 0, rbt_len = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("%s block_addr %d\n",__FUNCTION__,block_addr);
    /*get free current rba number -->use different logical block*/
    while(this->cur_rba_num > 0)
    {
        /*reduce rba number in the memory*/
        remap_block  = sn_get_remap_block(mtd, block_addr);
        printf("%s reduce block_addr %d remap_block is %d cur rab num %d \n",
			__FUNCTION__,block_addr,remap_block,this->cur_rba_num);
        block_addr++;
    }
    dump_sn_bbt(mtd);
	printf("%s log block %d phyblock %d \n",__FUNCTION__,
		block_addr,sn_get_log2phy_block(mtd,block_addr));

	printf("%s start rbt mode \n",__FUNCTION__);
    rbt_len = this->bbt_len;
	while(rbt_len > 0)
    {
        /*reduce rba in the memory*/
        remap_block  = sn_get_free_block(mtd, block_addr);
        printf("%s reduce block_addr is %d remap_block %d \n",
			__FUNCTION__,block_addr,remap_block);
        block_addr++;
        rbt_len--;
    }
    dump_sn_rbt(mtd);
	printf("%s log block %d phyblock %d \n",__FUNCTION__,
		block_addr,sn_get_log2phy_block(mtd,block_addr));
	
    remap_block = sn_get_free_block(mtd,block_addr);
    printf("%s remap_block is %d cur rab num %d \n",
			__FUNCTION__,remap_block,this->cur_rba_num);
    dump_sn_bbt(mtd);
    dump_sn_rbt(mtd);
	printf("%s log block %d phyblock %d \n",__FUNCTION__,
		block_addr,sn_get_log2phy_block(mtd,block_addr));
}

void sn_test_erase_write(struct mtd_info *mtd, int block_addr)
{
#ifdef CONFIG_MTD_TABLE_TEST
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	struct erase_info *instr;
	unsigned long long to = 0;
	unsigned int  retlen = 0;
	u_char* write_buffer = NULL, *read_buffer = NULL;
	int loop = 0,page_id = 0,offset = 0;
	instr->mtd = mtd;
	instr->addr = block_addr*this->block_size;
	instr->len = this->block_size;
	to = block_addr*this->block_size;
	if((write_buffer =	kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) ==	NULL){
		printf("%s malloc write_buffer error! \n",__FUNCTION__);
		return;
	}
	
	if((read_buffer =  kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) ==  NULL){
		printf("%s malloc read_buf error!\n",__FUNCTION__);
		kfree(write_buffer);
		return;
	}
	printf("%s block_addr %d \n",__FUNCTION__,block_addr);
	get_random_content(write_buffer,this->block_size);
	if(block_addr < this->bootcode_end_phy_block){
		printf("this block_addr %d not kernel block %d \n",block_addr,this->bootcode_end_phy_block);
		return;
	}else{
		for(loop = 0; loop < 2 ; loop++)
		{
			/*make erase fail:times:2*/
			set_erase_fail = 1;
			printf("%s loop %d set_erase_fail %d \n",__FUNCTION__,loop,set_erase_fail);
			mtd->erase(mtd, instr);
			for(page_id = 0; page_id < this->block_per_page; page_id++)
			{
				/*make write fail:times: this->block_per_page*2*/
				offset = page_id*this->page_size;
				set_write_fail = 1; 
				if(mtd->write_ecc(mtd, to+offset, this->page_size, &retlen, (u_char *)write_buffer+offset,
					NULL, NULL)!= 0){
					printf("%s write fail page id %d\n", __FUNCTION__,page_id);
					break;
				}
				
				mtd->read_ecc(mtd, to+offset, this->page_size, &retlen, (u_char *)read_buffer+offset, NULL, NULL);
				if(memcmp(write_buffer+offset,read_buffer+offset,this->page_size) == 0){
					printf("compare page_id %d offset 0x%x write read buffer ok! \n \n",
						page_id,offset);
				}else{
					printf("compare page_id %d offset 0x%x write read buffer fail! \n",
						page_id,offset);
					dump_byte(write_buffer+offset,this->page_size);
					dump_byte(read_buffer+offset,this->page_size);
					break;
				}	
			}
		}
	}
	
	if(read_buffer!= NULL){
		kfree(read_buffer);
	}
	
	if(write_buffer!= NULL){
		kfree(write_buffer);
	}
#endif
}

void sn_test_read(struct mtd_info *mtd,int block_addr)
{
#ifdef CONFIG_MTD_TABLE_TEST
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	struct erase_info *instr;
	unsigned long long to = 0;
	unsigned int  retlen = 0;
	u_char* write_buffer = NULL,*read_buffer = NULL;
	int page_id = 0,offset = 0;
	instr->mtd = mtd;
	instr->len = this->block_size;

	if((write_buffer =	kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) == NULL){
		printf("%s malloc write_buffer error!\n",__FUNCTION__);
		return;
	}

	if((read_buffer =  kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) ==  NULL){
		printf("%s malloc read_buf error!\n",__FUNCTION__);
		kfree(write_buffer);
		return;
	}

	instr->addr = block_addr*this->block_size;
	if(mtd->erase(mtd, instr) < 0){
		printf("%s erase block %d error!\n",__FUNCTION__,block_addr);
		return;
	}
	get_random_content(write_buffer,this->block_size);
	to = block_addr*this->block_size;
	for(page_id = 0; page_id < this->block_per_page; page_id++)
	{
		offset = page_id*this->page_size;
		if(mtd->write_ecc(mtd, to+offset, this->page_size,&retlen,
				(u_char *)write_buffer+offset,NULL, NULL)!= 0){
			printf("%s write fail page id %d\n", __FUNCTION__,page_id);
			break;
		}

		set_read_fail = 1;
		/*read next page ,recovery the current page*/
		mtd->read_ecc(mtd, to+offset+this->page_size, this->page_size,
			&retlen,(u_char *)read_buffer+offset, NULL, NULL);

		/*read current page for test*/
		if(mtd->read_ecc(mtd, to+offset, this->page_size,
			&retlen,(u_char *)read_buffer+offset, NULL, NULL)!=0 ){
			printf("%s read fail page id %d\n", __FUNCTION__,page_id);
			break;
		}
		
		if(memcmp(write_buffer+offset,read_buffer+offset,this->page_size) == 0){
			printf("compare page_id %d offset 0x%x write read buffer ok! \n \n",
						page_id,offset);
		}else{
			printf("compare page_id %d offset 0x%x write read buffer fail! \n",
				page_id,offset);
			dump_byte(write_buffer+offset,this->page_size);
			dump_byte(read_buffer+offset,this->page_size);
			break;
		}
	}
	if(read_buffer!= NULL){
		kfree(read_buffer);
	}
	
	if(write_buffer!= NULL){
		kfree(write_buffer);
	}
#endif
}


void sn_test_no_collect_table(struct mtd_info *mtd,int block_addr,int way)
{
#ifdef CONFIG_MTD_TABLE_TEST
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	struct erase_info *instr;
	unsigned long long to = 0;
	unsigned int  retlen = 0;
	u_char* write_buffer = NULL;
	int page_id = 0,offset = 0,remap_block = 0,bad_block = 0;
	instr->mtd = mtd;
	instr->len = this->block_size;
	instr->addr = block_addr*this->block_size;
	to = block_addr*this->block_size;
	if((write_buffer =	kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) ==	NULL){
		printf("%s malloc write_buffer error! \n",__FUNCTION__);
		return;
	}

	bad_block = block_addr;
	while(this->cur_rba_num > 0)
	{
		/*reduce rba in the memory*/
		remap_block  = sn_get_free_bbt_block(mtd, bad_block);
		if(way != 0){
			bad_block++;
		}
		printf("%s %d reduce block_add %d remap_block is %d free rab num %d cur rba num %d \n",
			__FUNCTION__,__LINE__,block_addr,remap_block,this->cur_rba_num,this->use_rba_num);
	}

	if(way == 0){
		dump_sn_bbt(mtd);
	}else{
		dump_sn_rbt(mtd);
	}

	/*no free block -->use collect block or rbt*/
	if(mtd->erase(mtd, instr) < 0){
		printf("%s erase block %d error! \n",__FUNCTION__,remap_block);
		return;
	}

	get_random_content(write_buffer,this->block_size);
	for(page_id = 0; page_id < this->block_per_page; page_id++)
	{
		offset = page_id*this->page_size;
		/*make all collect block(this->bbt_len + 1),include bbt region(this->bi_table_block_num)+10
		  erase ok, write fail case, so no block to write ,so need break case*/
		set_write_fail= this->bbt_len + 1 +this->bi_table_block_num + 10; 
		printf("%s entry mark bad block %d use next block!\n",__FUNCTION__,block_addr);
		if(mtd->write_ecc(mtd, to+offset, this->page_size, &retlen, 
			(u_char *)write_buffer+offset,NULL, NULL)!= 0){
			printf("%s write fail page id %d\n", __FUNCTION__,page_id);
			break;
		}	
	}

	if(write_buffer!= NULL){
		kfree(write_buffer);
	}
#endif
}

void sn_test_bi_table(struct mtd_info *mtd,int test_mode)
{
#ifdef CONFIG_MTD_TABLE_TEST
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	if(test_mode == 0){
		printf("read frist block error,entry backup block,and recovery:"
			"frist block write ok,recover use read fail block!\n");
		set_read_fail = 1;
		load_block_table_from_sn(mtd,0xBB,0);
		dump_sn_bi_table(mtd);
		
		printf("read frist block error,entry backup block, and recovery: "
			"frist block write fail,so recover free block\n");
		set_read_fail = 1;
		set_write_fail= 1;
		load_block_table_from_sn(mtd,0xBB,0);
		dump_sn_bi_table(mtd);
	}else{
		printf("write frist block error,entry free block way\n");
		set_write_fail = 1;
		update_bbt_to_sn(mtd,(uchar *)this->bbt_table,(sizeof(bbt_info_t)*this->bbt_len));
		dump_sn_bi_table(mtd);
	}
#endif
} 

static void sn_log2phy_test(struct mtd_info *mtd,int block_addr)
{
    int phy_block = sn_get_log2phy_block(mtd,block_addr);
    printf("%s log %d phy %d \n",__FUNCTION__,block_addr,phy_block);
}

static void sn_table_test(struct mtd_info *mtd,int block_addr,int test_way)
{
    if(test_way == 0)
    {
        sn_bbt_collect_test(mtd,block_addr); 
    }else if(test_way == 1){
	    sn_rbt_collect_test(mtd,block_addr);
    }
}

void sn_erase_perf_test(struct mtd_info *mtd,unsigned int start_addr,unsigned int size)
{
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	unsigned long begin = 0,end = 0,spend_time = 0;
	unsigned int block_id = 0, block_start = 0,block_num = 0,rate = 0;
	block_start = start_addr / this->block_size;
	block_num   = size/this->block_size;
	WATCHDOG_RESET();	
	begin = rtd_inl(0x1801B63C);    
    
    for(block_id = 0; block_id < block_num; block_id++)
    {
        this->erase_block(mtd,block_start+block_id);
    }
	
	end  = rtd_inl(0x1801B63C);
    spend_time =  (end-begin)/27;
	rate = spend_time/block_num;
	printf("%s erase time begin %lu end %lu time %luus size 0x%x erase block time %5dus \n",
		__FUNCTION__,begin,end,spend_time,size,rate);
	
}

void sn_write_perf_test(struct mtd_info *mtd,unsigned int start_addr,unsigned int size)
{
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	unsigned long begin = 0,end = 0,spend_time = 0;
	unsigned int loop = 0,page_num = 0,page_start = 0,rate = 0;
	unsigned char *page_content = (unsigned char *)malloc(1<<this->page_shift);
	if(page_content == NULL){
		printf("malloc size 0x%x fail!\n",this->page_size);
		return;
	}
	get_random_content(page_content,this->page_size);
	page_start = start_addr/this->page_size;
	page_num = size/this->page_size;
	WATCHDOG_RESET();	
	begin = rtd_inl(0x1801B63C);   
	for(loop = 0; loop < page_num; loop++)
    {
        if(this->write_ecc_page(mtd,page_start+loop,page_content,this->g_oobbuf)!=0){
            break;
        }
    }
	end  = rtd_inl(0x1801B63C);
    spend_time =  (end-begin)/27;
	rate = (size/1024)*1000000/spend_time;
	printf("%s write time begin %lu end %lu time %luus size 0x%x %5dKB/s\n",
		__FUNCTION__,begin,end,spend_time,size,rate);
	if(page_content != NULL){
		free(page_content);
	}
}

void sn_read_perf_test(struct mtd_info *mtd,unsigned int start_addr,unsigned int size)
{
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	unsigned long begin = 0,end = 0,spend_time = 0;
	unsigned int loop = 0,page_num = 0,page_start = 0,rate = 0;
	unsigned char *page_content = (unsigned char *)malloc(1<<this->page_shift);
	if(page_content == NULL){
		printf("malloc size 0x%x fail!\n",this->page_size);
		return;
	}
	page_start = start_addr/this->page_size;
	page_num = size/this->page_size;
	
	read_measure = 1;
	WATCHDOG_RESET();	
	begin = rtd_inl(0x1801B63C);  
	for(loop = 0; loop < page_num;loop++)
	{
		dump_by_cpu(mtd,page_start+loop);
	}
	end  = rtd_inl(0x1801B63C);
    spend_time =  (end-begin)/27;
	rate = (size/1024)*1000000/spend_time;
	printf("%s read cpu begin %lu end %lu time %luus size 0x%x %6dKB/s \n",
		__FUNCTION__,begin,end,spend_time,size,rate);
	read_measure = 0;
	
	WATCHDOG_RESET();	
	begin = rtd_inl(0x1801B63C);  
	for(loop = 0; loop < page_num;loop++)
	{
		spi_nand_read_page(mtd,page_start+loop,1,page_content);
	}
	end  = rtd_inl(0x1801B63C);
    spend_time =  (end-begin)/27;
	rate = (size/1024)*1000000/spend_time;
	printf("%s read md begin %lu end %lu time %luus size 0x%x %6dKB/s\n",
		__FUNCTION__,begin,end,spend_time,size,rate);
	
	if(page_content != NULL){
		free(page_content);
	}
}

void sn_performace_test(struct mtd_info *mtd)
{
	unsigned int start_addr = 0, size = 0;
	start_addr = SYS_PARAM_FACTORY_RO_START;
	size	   = SYS_PARAM_FACTORY_RO_SIZE;
	printf("%s start address 0x%x size 0x%x \n",__FUNCTION__,start_addr,size);
	/*erase performace test*/
	sn_erase_perf_test(mtd,start_addr,size);
	
	/*write performace test*/
	sn_write_perf_test(mtd,start_addr,size);

	/*read performace test*/
	sn_read_perf_test(mtd,start_addr,size);
}

int do_spi_nand(cmd_tbl_t * cmdtp, int flag, int argc, char * const argv[])
{
    int loop = 0, dev = spi_nand_curr_device;
    struct mtd_info *spi_nand = &spi_nand_info[dev];
    ulong addr = 0, size = 0,off = 0;
	char *cmd = NULL;
	  
    /* at least two arguments please */
    if (argc < 2)
    {
        printf("%s arguments %d > 2 eg: arg0:spi_nand arg1:cmd(/info/bad/status/..)\n",
        __FUNCTION__,argc);
        goto usage;
    }

    cmd = argv[1];
    if (strcmp(cmd, "reinit") == 0) 
    {
        spi_nand_init();
		return 0;
    }

    if (strcmp(cmd, "info") == 0) 
    {
        putc('\n');
		for (loop = 0; loop < CONFIG_SYS_MAX_SPI_NAND_DEVICE; loop++) 
        {
            if (spi_nand_info[loop].name)
                spi_nand_print_and_set_info(loop);
        }
        return 0;
    }

    if (strcmp(cmd, "bad") == 0) 
    {
        scan_spi_nand_block_status(spi_nand);
        return 0;
    }

    if (strcmp(cmd, "status") == 0)
    {
        spi_nand_get_bootcode_status(spi_nand);
        return 0;
    }

	
		
    if (strcmp(cmd, "pref_test") == 0)
    {
    	sn_performace_test(spi_nand);
        return 0;
    }

	
    if (strcmp(cmd, "dump") == 0)
    {
        if (argc > 3)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 16);
            size = (ulong)simple_strtoul(argv[3], NULL, 16);        
            dump_content(spi_nand,addr,size,0,NULL);
        }
        else
        {
            printf("arg == 4,arg2:start address(hex) arg3: size(hex)\n");
            goto usage;
        }
        return 0;
    }


    if (strcmp(cmd, "dump_raw") == 0)
    {
        if (argc > 3)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 16);
            size = (ulong)simple_strtoul(argv[3], NULL, 16);        
            dump_content(spi_nand,addr,size,1,NULL);
        }
        else
        {
            printf("arg == 4,arg2:start address(hex) arg3: size(hex)\n");
            goto usage;
        }
        return 0;
    }
	
	if (strcmp(cmd, "dump_cpu") == 0)
    {
        if (argc > 2)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);       
            dump_by_cpu(spi_nand,addr);
            return 0;
        }
        else
        {
            printf("arg == 3,arg2:page address\n");
            goto usage;
        }
    }

	if (strcmp(cmd, "dump_ddr") == 0)
	{
		if (argc > 4)
		{
			addr = (ulong)simple_strtoul(argv[2], NULL, 10);
			size = (ulong)simple_strtoul(argv[3], NULL, 16);    
			off  = (ulong)simple_strtoul(argv[4], NULL, 16); 
			dump_content(spi_nand,addr,size,0,(char *)off);
			return 0;
		}
		else
		{
			printf("arg == 4,arg2:start address(hex) arg3: size(hex),arg4:ddr (hex)\n");
			goto usage;
		}
	}
	
	
    if (strcmp(cmd, "dump_oob") == 0)
    {
        if (argc > 2)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);       
            dump_page_oob(spi_nand,addr);
            return 0;
        }
        else
        {
            printf("arg == 3,arg2:page address\n");
            goto usage;
        }
    }

    if (strcmp(cmd, "erase") == 0)
    {
        if (argc > 3)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);  
            size = (ulong)simple_strtoul(argv[3], NULL, 10); 
            erase_spi_nand_block(spi_nand,addr,size);
            return 0;
        }
        else
        {
            printf("arg == 4,arg2:block address arg2:block number\n");
            goto usage;
        }
    }

    if (strcmp(cmd, "write") == 0)
    {
        if (argc > 4)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 16);
            size = (ulong)simple_strtoul(argv[3], NULL, 16);  
            off  = (ulong)simple_strtoul(argv[4], NULL, 16);   
            spi_nand_write_random(spi_nand,(char *)off,addr,size,0);
            return 0;
        }
        else
        {
            printf("arg == 4,arg2:start address(hex) arg3: size(hex),arg4:buffer(hex)\n");
            goto usage;
        }
    }

    if (strcmp(cmd, "write_oob") == 0)
    {
        if (argc > 3)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);
            size = (ulong)simple_strtoul(argv[3], NULL, 16);  
            write_oob(spi_nand,addr,size);
            return 0;
        }
        else
        {
            printf("arg == 4,arg2:page address arg3: oob content\n");
            goto usage;
        }
    }

    
    if (strcmp(cmd, "mark_bad") == 0)
    {
        if (argc > 2)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);  
            mark_badblock(spi_nand,addr);
            return 0;
        }
        else
        {
            printf("arg == 3,arg2:block address\n");
            goto usage;
        }
    }

	if (strcmp(cmd, "cache_mode") == 0)
    {
        if (argc > 2)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);       
            set_cache_mode(spi_nand,addr);
            return 0;
        }
        else
        {
            printf("arg == 3,arg2:cache mode\n");
            goto usage;
        }
    }	

    if (strcmp(cmd, "wp_test") == 0)
    {
        if (argc > 3)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);   
            size = (ulong)simple_strtoul(argv[3], NULL, 10);  
            wp_block_test(spi_nand,addr,size);
            return 0;
        }
        else
        {
            printf("arg == 4,arg2:block address arg3:block number\n");
            goto usage;
        }
    }	

	if(strcmp(cmd, "dump_bi") == 0)
    {
        dump_sn_bi_table(spi_nand);
        return 0;
    }

    if(strcmp(cmd, "dump_ibt") == 0)
    {
        dump_sn_ibt(spi_nand);
        return 0;
    }

    if(strcmp(cmd, "dump_bbt") == 0)
    {
        dump_sn_bbt(spi_nand);
        return 0;
    }

	if(strcmp(cmd, "dump_rbt") == 0)
    {
        dump_sn_rbt(spi_nand);
        return 0;
    }

    if(strcmp(cmd, "dump_log2phy") == 0)
    {
        dump_sn_log2phy_talbe(spi_nand);
        return 0;
    }

	
    if (strcmp(cmd, "log2phy_test") == 0)
    {
        if (argc > 2)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);    
            sn_log2phy_test(spi_nand,addr);
            return 0;
        }
        else
        {
            printf("arg == 3,arg2:block address\n");
            goto usage;
        }
    }

    if (strcmp(cmd, "table_test") == 0)
    {
        if (argc > 3)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);   
            size = (ulong)simple_strtoul(argv[3], NULL, 10);  
            sn_table_test(spi_nand,addr,size);
            return 0;
        }
        else
        {
            printf("arg == 4,arg3:table_test way arg2:block address\n");
            goto usage;
        }
    }

    if (strcmp(cmd, "table_wp_test") == 0)
    {
        if (argc > 2)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);       
            table_wp_test(spi_nand,addr);
            return 0;
        }
        else
        {
            printf("arg == 3,arg2:block address\n");
            goto usage;
        }
    }	
	
    if(strcmp(cmd, "dump_reg") == 0)
    {
	    dump_sn_reg();
	    return 0;
    }

	if (strcmp(cmd, "bi_test") == 0)
    {
        if (argc > 2)
        {
            addr = (ulong)simple_strtoul(argv[2], NULL, 10);       
            sn_test_bi_table(spi_nand,addr);
            return 0;
        }
        else
        {
            printf("arg == 3,arg2:test way!\n");
            goto usage;
        }
    }	

	if (strcmp(cmd, "ewr_test") == 0) {
		if (argc < 4) {
			printf("arg == 4,arg2:address!\n");
			goto usage;
		}
		addr = 0;
		addr = (int)simple_strtoul(argv[2], NULL, 10);
		size = (int)simple_strtoul(argv[3], NULL, 10);
	
		if(size == 0){
			sn_test_erase_write(spi_nand, addr);
		}else if(size == 1){
			sn_test_read(spi_nand, addr);
		}
		return 0;
	}

    if (strcmp(cmd, "no_col_table") == 0) {
		if (argc < 4) {
			printf("arg == 4,arg2:address agr3:test way!\n");
			goto usage;
		}
		addr = (int)simple_strtoul(argv[2], NULL, 10);
		size	= (int)simple_strtoul(argv[3], NULL, 10);
		sn_test_no_collect_table(spi_nand,addr,size);
		return 0;
	}

usage:
	return CMD_RET_USAGE;
}

U_BOOT_CMD(
    spi_nand, CONFIG_SYS_MAXARGS, 1, do_spi_nand,
    "SPI_NAND sub-system",
    "spi_nand reinit                            - re-initialize spi_nand driver\n"
    "spi_nand info                              - show available SPI NAND devices\n"
    "spi_nand bad                               - show all block bad or good\n"
    "spi_nand status                            - show bootcode block tag\n"
    "spi_nand pref_test                         - performace test \n"
    "spi_nand dump  address   size              - dump logical address(hex) size(hex) \n"
    "spi_nand dump_raw address size             - dump physical address(hex) size(hex) \n"
    "spi_nand dump_cpu page_addr                - dump page content by cpu way\n"  
    "spi_nand dump_oob page_addr                - dump page oob content\n"
    "spi_nand dump_ddr nand_addr size ddr_addr  - dump address content to ddr\n"
    "spi_nand erase address size                - erase block address size\n"
    "spi_nand write nand_addr size ddr_addr     - write address(hex) size(hex) buf(hex)\n"
    "spi_nand write_oob address size            - write oob page address spare(hex)\n"
    "spi_nand mark_bad address                  - mark bad block address \n"
    "spi_nand cache_mode mode                   - set spi_nand cache mode 0:normal 1:dual 2:quard 3:quad_IO \n"
    "spi_nand wp_test blockid  size             - wp test blockid:physical block id size:block number wp times\n"
    "spi_nand log2phy_test address              - log2phy table address:block \n"
    "spi_nand table_test address size           - table_test address:block size:0 bbt size:1 rbt\n"
    "spi_nand table_wp_test blockid             - wp test use bbt talbe,logical block address\n"
    "spi_nand bi_test  test_mode                - test_mode:0(loader) 1(write)\n"
	"spi_nand ewr_test block_addr  test_way     - block_addr:block id test_way:0:ew 1:r\n"
	"spi_nand no_col_table block_addr  test_way - block_addr:block id test_way:0:bbt 1:rbt\n"
    "spi_nand dump_bi                           - show spi_nand bi table content\n"
    "spi_nand dump_ibt                          - show spi_nand ibt content\n"
    "spi_nand dump_bbt                          - show spi_nand bbt content\n" 
    "spi_nand dump_rbt                          - show spi_nand rbt content\n" 
    "spi_nand dump_log2phy                      - show spi_nand log2phy table content\n"
    "spi_nand dump_reg                          - show spi_nand register\n"
);
